#include <float.h>
#include <stdio.h>
#include <math.h>

#include <mkl.h>

#include "dot.h"
#include "timer.h"

#include <xmmintrin.h>

//static assures that this function will only be used inside of this file.
//This gives the compiler more fredom w.r.t. optimizations (i.e. inlining)
static float dot4_v1(float * restrict x, float * restrict y); 

static float dot4_v2(float * restrict x, float * restrict y);

static float dot4_v3(float * restrict x, float * restrict y);

static float dot4_v4(float * restrict x, float * restrict y);

float dot4(float *x, float *y )
{

   /*****************************************
    * Validation 
    *****************************************/
   float mkl_result = cblas_sdsdot (4, 0.0f, x, 1, y, 1);

   float rel_difference1 = fabs( (mkl_result - dot4_v1(x,y)) / mkl_result);
   float rel_difference2 = fabs( (mkl_result - dot4_v2(x,y)) / mkl_result);
   float rel_difference3 = fabs( (mkl_result - dot4_v3(x,y)) / mkl_result);
   float rel_difference4 = fabs( (mkl_result - dot4_v4(x,y)) / mkl_result);
   if(rel_difference1 < 1e-4)
      printf("Unit test 1 \033[32m SUCCESSFUL!\033[0m\n");
   else{
      printf("Unit test 1 \033[31m FAILED!\033[0m\n");
   }
   if(rel_difference2 < 1e-4)
      printf("Unit test 2 \033[32;1m SUCCESSFUL!\033[0m\n");
   else{
      printf("Unit test 2 \033[31m FAILED!\033[0m\n");
   }
   if(rel_difference3 < 1e-4)
      printf("Unit test 3 \033[32;1m SUCCESSFUL!\033[0m\n");
   else{
      printf("Unit test 3 \033[31m FAILED!\033[0m\n");
   }
   if(rel_difference4 < 1e-4)
      printf("Unit test 4 \033[32;1m SUCCESSFUL!\033[0m\n");
   else{
      printf("Unit test 4 \033[31m FAILED!\033[0m\n");
   }
   /****************************************/


   float result = 0.0f;

   double min_time = FLT_MAX;

   for(int j=0;j < 1000; ++j){

      unsigned long start, end;

      get_ticks(start);
      for(int i=0;i < 10000; ++i){

         result += dot4_v4(x,y);

      }

      get_ticks(end);
      double elapsed = ((double)(end-start))/10000;
      if( elapsed < min_time )
         min_time = elapsed;

      
   }

   printf("Cycles per function call: %g\n",min_time);

   return result;
}


float dot4_v1(float *x, float *y)
{
   return 0;
}



float dot4_v2(float *x, float *y)
{
   return 0;
}



float dot4_v3(float *x, float *y)
{
   __m128 x_xmm = _mm_loadu_ps(x);
   __m128 y_xmm = _mm_loadu_ps(y);

   x_xmm = _mm_mul_ps(x_xmm, y_xmm);

   y_xmm = _mm_movehl_ps(x_xmm, x_xmm);

   y_xmm = _mm_add_ps(x_xmm, y_xmm);
   x_xmm = _mm_movehdup_ps(y_xmm);
   x_xmm = _mm_add_ps(x_xmm, y_xmm);

   float result __attribute__((aligned(32)));
   _mm_store_ss(&result, x_xmm); 

   return result;
}


float dot4_v4(float *x, float *y)
{
   __m128 x_xmm = _mm_loadu_ps(x);
   __m128 y_xmm = _mm_loadu_ps(y);

   x_xmm = _mm_mul_ps(x_xmm, y_xmm);

   y_xmm = _mm_permute_ps(x_xmm, 0xEE);

   y_xmm = _mm_add_ps(x_xmm, y_xmm);
   x_xmm = _mm_permute_ps(y_xmm, 0x1);
   x_xmm = _mm_add_ps(x_xmm, y_xmm);

   float result __attribute__((aligned(32)));
   _mm_store_ss(&result, x_xmm); 

   return result;
}














/*
float dot4_v1(float *x, float *y)
{
   __m128 x_xmm = _mm_loadu_ps(x);
   __m128 y_xmm = _mm_loadu_ps(y);

   x_xmm = _mm_dp_ps(x_xmm, y_xmm, 0xFF);

   float result __attribute__((aligned(32)));
   _mm_store_ss(&result, x_xmm); 

   return result;
}


float dot4_v2(float *x, float *y)
{
   __m128 x_xmm = _mm_loadu_ps(x);
   __m128 y_xmm = _mm_loadu_ps(y);

   x_xmm = _mm_mul_ps(x_xmm, y_xmm);

   x_xmm = _mm_hadd_ps(x_xmm, x_xmm);
   x_xmm = _mm_hadd_ps(x_xmm, x_xmm);

   float result __attribute__((aligned(32)));
   _mm_store_ss(&result, x_xmm); 

   return result;
}





float dot4_v3(float *x, float *y)
{
   __m128 x_xmm = _mm_loadu_ps(x);
   __m128 y_xmm = _mm_loadu_ps(y);

   x_xmm = _mm_mul_ps(x_xmm, y_xmm);

   y_xmm = _mm_movehl_ps(x_xmm, x_xmm);

   y_xmm = _mm_add_ps(x_xmm, y_xmm);
   x_xmm = _mm_movehdup_ps(y_xmm);
   x_xmm = _mm_add_ps(x_xmm, y_xmm);

   float result __attribute__((aligned(32)));
   _mm_store_ss(&result, x_xmm); 

   return result;
}

float dot4_v4(float *x, float *y)
{
   __m128 x_xmm = _mm_loadu_ps(x);
   __m128 y_xmm = _mm_loadu_ps(y);

   x_xmm = _mm_mul_ps(x_xmm, y_xmm);

   y_xmm = _mm_permute_ps(x_xmm, 0xEE);

   y_xmm = _mm_add_ps(x_xmm, y_xmm);
   x_xmm = _mm_permute_ps(y_xmm, 0x1);
   x_xmm = _mm_add_ps(x_xmm, y_xmm);

   float result __attribute__((aligned(32)));
   _mm_store_ss(&result, x_xmm); 

   return result;
}

*/

